#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
ebooted - EBOOT editor and (un)packer

Copyright (c) 2009, Tobias Theuer <tobi at unhol dot de>

Permission to use, copy, modify, and/or distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright notice
and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
THIS SOFTWARE.
"""

##### ----- imports ----- ######
import sys, os, struct, tempfile
from optparse import OptionParser, OptionGroup
import filesize


##### ----- global variables ----- ######
program_name = "ebooted - EBOOT editor (un)packer"
script_name = os.path.basename(sys.argv[0])

version_major = 0
version_minor = 1
version_revision = 1
version = str(version_major) + "." + str(version_minor) + "." + str(version_revision)
version_message = program_name + " " + version
#define version & version message

usage = script_name + " -f FILE [options] [files]"
#define help-message

overwrite = False
#overwrite existing files?

verbose = False
quick = False
#TODO: Quickmode not yet implemented. Might do that someday.

temp_dir = None

##### ----- class in_file ----- ######
class in_file:
    """Files inside the PBP-file, including filename, filesize (and more?)"""
    
    ### --- def __init__ --- ###
    def __init__(self, filename, offset, size = 0):
        self.name = filename
        self.offset = offset
        self.size = size
    
    ### --- def to_string --- ###      
    def to_string(self):
        if verbose:
            return "%-9s %15s" % (self.name, str(self.size) + " bytes")
        else:
            return "%-9s %10s" % (self.name, filesize.size(self.size, filesize.alternative))


##### ----- class pbp ----- ######
class pbp:
    """The main class of the program. Reading/Writing PBP-files"""
    
    ### --- def __init__ --- ###
    def __init__(self, filename = None):
        
        if filename:
            self.filename = filename
            self.file_size = os.path.getsize(filename)
            self.file_handler = open(filename, "rb+")
            self.signature, self.offset_param_sfo, self.offset_icon0_png, self.offset_icon1_pmf, self.offset_pic0_png, self.offset_pic1_png, self.offset_snd0_at3, self.offset_data_psp, self.offset_data_psar = struct.unpack("4sxxxx8I", self.file_handler.read(40))
            # 4s = string of length 4; x = pad byte (ignored); I = unsigned int
                    
            if self.signature != "\x00PBP":
                raise Exception("file " + filename + " is not a compatible BPB file")

            self.file_list = []
        
            self.file_list.append(in_file("param.sfo", self.offset_param_sfo))
            if self.offset_icon0_png != self.offset_icon1_pmf:
                self.file_list.append(in_file("icon0.png", self.offset_icon0_png))
            if self.offset_icon1_pmf != self.offset_pic0_png:
                self.file_list.append(in_file("icon1.pmf", self.offset_icon1_pmf))
            if self.offset_pic0_png != self.offset_pic1_png:
                self.file_list.append(in_file("pic0.png", self.offset_pic0_png))
            if self.offset_pic1_png != self.offset_snd0_at3:
                self.file_list.append(in_file("pic1.png", self.offset_pic1_png))
            if self.offset_snd0_at3 != self.offset_data_psp:
                self.file_list.append(in_file("snd0.at3", self.offset_snd0_at3))
            if self.offset_data_psp != self.offset_data_psar:
                self.file_list.append(in_file("data.psp", self.offset_data_psp))
            if self.offset_data_psar != self.file_size:
                self.file_list.append(in_file("data.psar", self.offset_data_psar))
            
            self.file_list.append(in_file("dummy", self.file_size)) 
            #to make the loop below work correctly
            
            for i in xrange(len(self.file_list)-1):
                self.file_list[i].size = self.file_list[i+1].offset - self.file_list[i].offset
            #go through the list and calculate the file-sizes
            
            del self.file_list[-1]
            #remove the dummy list entry    
            
        else:
        #no filename was supplied, so nothing will be loaded, just empty variables to declare
            self.file_handler, self.file_size, self.offset_param_sfo, self.offset_icon0_png, self.offset_icon1_pmf, self.offset_pic0_png, self.offset_pic1_png, self.offset_snd0_at3, self.offset_data_psp, self.offset_data_psar = 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
            self.filename, self.signature = "", ""
            self.file_list = []
                
    ### --- def list_files --- ###
    def list_files(self):
        this_list = []
        for item in self.file_list:
            this_list.append(item.to_string())
        return "\n".join(this_list)

    ### --- def unpack_all --- ###
    def unpack_all(self, directory):
        this_list = []
        for item in self.file_list:
            this_list.append(item.name)
        self.unpack_files(directory, this_list)

    ### --- def unpack_files --- ###
    def unpack_files(self, directory, files):
        if directory:
            if not os.path.exists(directory):
                os.makedirs(directory)
                #if directory doesn't exist, make it
                if verbose:
                    print "creating directory: " + directory
            if not os.path.isdir(directory):
                raise Exception(directory + " is not a directory or not accessible")
            path = directory
        else:
            path = ""
        
        to_unpack_list = []
        existing_files = []
                
        for arg_file in files:
            for existing_file in self.file_list:
                if arg_file.lower() == existing_file.name:
                    to_unpack_list.append(existing_file)
                    existing_files.append(existing_file.name)
                    break
        
        for arg_file in files:
            if arg_file not in existing_files:
                raise Exception(arg_file + " not found in " + self.filename + ". Use --list option to see available files.")
        
        if verbose:
            print "unpacking:"
            for item in to_unpack_list:
                print item.to_string()
            print
        
        for item in to_unpack_list:
            
            this_file = os.path.join(path, item.name)
                        
            if os.path.exists(this_file):
                if overwrite:
                    if verbose:
                        print "file " + this_file + " exists, overwriting."
                else:
                    raise Exception("file " + this_file + " exists. To overwrite files, use -w option.")
            
            self.file_handler.seek(item.offset, 0)
            #seek to the part of the in-file where the particular file to be unpacked is stored

            if verbose:
                print "writing " + this_file
                        
            out_file = open(this_file, "wb")
            #open file for writing
            
            file_buffer = 16777216
            file_size = item.size
            
            while file_size > 0:
                if file_size < file_buffer:
                    file_buffer = file_size
                                     
                out_file.write(self.file_handler.read(file_buffer))
                #read a buffer of bytes from in-file, write to out-file
                
                file_size -= file_buffer
            
            out_file.close()
    
            if verbose:
                print "success"
            
        if verbose:
            print "done"
    
    ### --- def pack --- ###
    def pack_files(self, target, directory, files):
        global quick
        
        if os.path.exists(target):
            if overwrite:
                if verbose:
                    print "file " + target + " exists, overwriting."
            else:
                raise Exception("file " + target + " exists. To overwrite it, use -w option.")
        
        allowed_files = ("param.sfo", "icon0.png", "icon1.pmf", "pic0.png", "pic1.png", "snd0.at3", "data.psp", "data.psar")
        
        these_files = {}

        if directory:
            if os.path.isdir(directory):
                these_files = dict([(i, os.path.join(directory, i)) for i in os.listdir(directory)])
            else:
                raise Exception(directory + " is not a directory or not accessible")
        #get a list of files (actually a dictionary with all files having their names as keys) in the directory (if it is given)
        
        if files:
            these_files.update(files)
        #merge the files from the directory (if any) with the files specified directly, overwriting in case of collisions
        
        for k, v in these_files.items():
            if not k in allowed_files:
                del these_files[k]
                #remove all files from the dictionary that should not be in a PBP-file
            else:
                if not os.path.exists(v):
                    raise Exception("file " + v + " not found")
                    #check if files exist
        
        if len(these_files) == 0:
            raise Exception("no files or no files belonging in a PBP-file were given")
        
        to_be_packed = {}
                
        for k, v in these_files.items():
            to_be_packed[k] = in_file(v, 0)
        #make a new dictionary of in_file-objects
        
        for v in to_be_packed.values():
            v.size = os.path.getsize(v.name)

        this_file_list = {}

        for i in self.file_list:
            this_file_list[i.name] = i
            
        """if quick:
            if len(self.file_list) > 0:
            #if we modify an existing file
                for k, v in to_be_packed.items():
                    for j, c in this_file_list.items():
                        if k == j:
                            if v.size > c.size:
                                quick = 0;
                                if verbose:
                                    print v.name + " bigger than " + k + " in " + target + ". Disabling quickmode."
                                break
                                #if the size of a file to be packed is bigger than that of a file in the existing pbp-file, quickmode (inserting instead of creating a new file) is impossible
        """
        #TODO: Quickmode
        
        this_file_list.update(to_be_packed)
        #merge the existing file-list with the new file-list
        
        if not 'param.sfo' in this_file_list and not 'data.psp' in this_file_list:
            raise Exception("No files designated for 'param.sfo' or 'data.psp'. Without these files, no working pbp-file can be created.")
    
        packing_list = []
        #another list, this time to hold the files to packed, in their correct order
        new_file_size = 0
        #the size of the new file
    
        for i in allowed_files:
            if i in this_file_list:
                packing_list.append(this_file_list[i])
                new_file_size += this_file_list[i].size
            else:
                packing_list.append(in_file(i, 0))
        #arrange the files in a sorted list and calculate the new file-size
        
        offset = 40
        #40 bytes reserved for the header
                
        for i in packing_list:
            if i.size == 0:
                i.offset = 0
            else:
                i.offset = offset
            offset += i.size
        #assign offsets for the existing files
        
        #now we go through the list backwards to assign offsets to the non-existing files in the package
        #this is necessary because, contrary to information on the web, non-existing files have the offset of the next existing file, not the previous
        
        packing_list.reverse()
        
        offset = new_file_size
        
        for i in packing_list:
            if i.offset == 0:
                i.offset = offset
            else:
                offset = i.offset
        
        packing_list.reverse()
            
        header = struct.pack("8s", "\x00PBP\x00\x00\01\00")
        #header with signature and version(?) included
            
            
        for i in packing_list:
            header += struct.pack("I", i.offset)
        #add the offsets to the header
        
        out_file = open(target, "wb")
        #open file for writing
        
        if verbose:
            print "writing header, 40 bytes"
        out_file.write(header)
        if verbose:
            print "success"
        
        
        for i in packing_list:
        
            if i.size == 0:
                if verbose:
                    print "skipping " + i.name + ", 0 bytes"
                continue

        
            if verbose:
                print "writing " + i.name + ", " + filesize.size(i.size, filesize.alternative)
            
            file_buffer = 16777216
            file_size = i.size
            file_in = open(i.name, "rb")
        
            while file_size > 0:
                if file_size < file_buffer:
                    file_buffer = file_size                         
                out_file.write(file_in.read(file_buffer))
                #read a buffer of bytes from in-file, write to out-file
                file_size -= file_buffer
                
            file_in.close()
            
            if verbose:
                print "success"
        
        out_file.close()

        if verbose:
            print "done"
        
    ### --- def __del__ --- ###
    def __del__(self):
        if self.file_handler:
            self.file_handler.close()
        
##### ----- MAIN ----- ######
def main():
    global overwrite, verbose, quick, temp_dir
    #declare global variables changed in this def

    option_parser = OptionParser(usage=usage, version=version_message)
    #begin parsing of command line options

    option_parser.set_defaults(packer_action = "list", overwrite = False, verbose = False, quick = True, temp_dir = None)
    #if no other action is given, we use "list" (as it is the only action that doesn't write any files)

    option_parser.add_option("-f", "--file", "--eboot", help = "the FILE (usually EBOOT.PBP) to read/write", metavar="FILE", action = "store", dest = "filename")
    #filename

    option_parser.add_option("-l", "--list", const = "list", help = "list the files packed in FILE (default action)", action = "store_const", dest = "packer_action")
    option_parser.add_option("-u", "--unpack", const = "unpack", help = "unpack the specified (or all) files in FILE. Use 'all' or specific filenames to unpack", action = "store_const", dest = "packer_action")
    option_parser.add_option("-p", "--pack", const = "pack", help = "pack the specified files (or all files in the specified directory) into FILE", action = "store_const", dest = "packer_action")
    #possible actions

    option_parser.add_option("-d", "--directory", dest = "directory", help = "directory to where to unpack to / directory which stores all the files to be packed with their proper names")
    #option_parser.add_option("-q", "--disable-quickmode", action = "store_false", dest = "quick", help = "DISABLE quickmode (when packing an existing FILE, quickmode might be faster)")
    #TODO: Quickmode not yet implemented. Might do that someday.
    option_parser.add_option("-w", "--overwrite", action = "store_true", dest = "overwrite", help = "overwrite existing files (!)")
    option_parser.add_option("-v", "--verbose", action = "store_true", dest = "verbose", help = "give detailed reports")
    #option_parser.add_option("-t", "--temp-dir", action = "store", dest = "temp_dir", metavar = "DIR", help = "specify a directory for storing temporary files, otherwise, the default temporary directory of your OS will be used.")
    #self-explaining options
    
    file_group = OptionGroup(option_parser, "Specific file options", "alternative/additionally to the --directory option you can (for unpacking) use 'all', or specify the files inside the FILE to pack")
    file_group.add_option("--param.sfo", dest = "param_sfo", metavar = "PARAM.SFO")
    file_group.add_option("--icon0.png", dest = "icon0_png", metavar = "ICON0.PNG")
    file_group.add_option("--icon1.pmf", dest = "icon1_pmf", metavar = "ICON1.PMF")
    file_group.add_option("--pic0.png", dest = "pic0_png", metavar = "PIC0.PNG")
    file_group.add_option("--pic1.png", dest = "pic1_png", metavar = "PIC1.PNG")
    file_group.add_option("--snd0.at3", dest = "snd0_at3", metavar = "SND0.AT3")
    file_group.add_option("--data.psp", dest = "data_psp", metavar = "DATA.PSP")
    file_group.add_option("--data.psar", dest = "data_psar", metavar = "DATA.PSAR")
    option_parser.add_option_group(file_group)
    #group the "sub-file"-specific options
    
    (options, args) = option_parser.parse_args()
    #end of parsing command line options, get them.

    overwrite = options.overwrite
    
    verbose = options.verbose
    
    #quick = options.quick
    #TODO: Quickmode not yet implemented. Might do that someday.

    #temp_dir = options.temp_dir

    filename = options.filename
    
    if not filename:
        option_parser.error("no file given")
        
    files = {}
    
    if options.param_sfo:
        files["param.sfo"] = options.param_sfo
    if options.icon0_png:
        files["icon0.png"] = options.icon0_png
    if options.icon1_pmf:
        files["icon1.pmf"] = options.icon1_pmf
    if options.pic0_png:
        files["pic0.png"] = options.pic0_png
    if options.pic1_png:
        files["pic1.png"] = options.pic1_png
    if options.snd0_at3:
        files["snd0.at3"] = options.snd0_at3
    if options.data_psp:
        files["data.psp"] = options.data_psp
    if options.data_psar:
        files["data.psar"] = options.data_psar
        
    if verbose:
        print "%9s: %10s" % ("Action", options.packer_action)
        if options.packer_action == "pack":
            print "%9s: %10s" % ("Target", filename)
        if options.packer_action == "list" or options.packer_action == "unpack" :
            print "%9s: %9s/" % ("File", filename)
            if options.directory:
                print "%9s: %9s/" % ("Target", options.directory)
            else:
                print "%9s: %9s/" % ("Target", ".")
        if options.packer_action != "list":            
            print "%9s: %10s" % ("Overwrite", overwrite)
        #if options.packer_action == "pack":
            #print "%9s: %10s" % ("Quickmode", quick)
            #TODO: Quickmode not yet implemented. Might do that someday.
            #if temp_dir:
            #    print "%9s: %10s" % ("Temp dir", temp_dir)
            #else:
            #    print "%9s: %10s" % ("Temp dir", tempfile.gettempdir())
            
        print
        #newline
        
    try:
        if options.packer_action != "pack":

            if not os.path.exists(filename):
                raise Exception("file " + filename + " not found")
                #does the file exist?
            file_handler = pbp(filename)

            if options.packer_action == "list":
                print file_handler.list_files()

            elif options.packer_action == "unpack":
                if len(args) < 1:
                    raise Exception("for unpacking you need to use either 'all' or specific files (see which files are in the file with --list)")
                else:
                    if args[0].lower() == "all":
                        file_handler.unpack_all(options.directory)
                    else:
                        file_handler.unpack_files(options.directory, args)

        else:
            if os.path.exists(filename):
                file_handler = pbp(filename)
            else:
                file_handler = pbp(None)
            #if the file to be packed already exists, i.e. an existing file is to be modified, read it, if not, don't

            file_handler.pack_files(filename, options.directory, files)

    except Exception, detail:
        print "Error:", detail.args[0:]
        sys.exit()

##### ----- PROGRAM ----- ######
if __name__ == "__main__":
    main()